from dreampylib import DreampyLib

from dnsprovider import DNSProvider


class DreamhostDNS(DNSProvider):
    def __init__(self, options):
        self.username = options['username']
        self.api_key = options['api_key']
        self.connection = None
        self._connect()

    def _connect(self):
        if self.connection and self.connection.IsConnected():
            return True

        self.connection = DreampyLib(self.username,
                                                self.api_key)

        return self.connection.IsConnected()

    def valid_response(self, response):
        try:
            return response and not (
                response[0] == False or response[1] == 'error')
        except:
            print response
            return False

    def add_record(self, data, hostName, type="A", TTL=3600, domainName=None):
        self._connect()

        # Dreamhost doesn't allow us to use TTLs
        return self.connection.dns.add_record(value=data,
                                              type=type,
                                              record=hostName)

    def remove_record(self, data, hostName, type="A", domainName=None):
        self._connect()

        return self.connection.dns.remove_record(record=hostName,
                                                 value=data,
                                                 type=type)

    def get_records(self, hostName=None, type="A", domainName=None):
        self._connect()

        records = self.connection.dns.list_records()

        if not hostName:
            return records

        records = [r for r in records if hostName in r['record']]

        if type == "*":
            return records

        return [r for r in records if r['type'] == type]



class MachineConfig(object):
    def __init__(self, hostname, shared_fate_zone,
                 cpus, mem, bits=64, disk=0, data=None,
                 ip=None):
        self.hostname = hostname
        self.cpus = cpus
        self.mem = mem
        self.shared_fate_zone = shared_fate_zone
        self.bits = bits
        self.disk = disk
        self.login_name = None
        self.login_key = None
        self.data = data
        self.dns_name = None
        self.ip = ip

    def __str__(self):
        return self.hostname

    def __repr__(self):
        return str(self)



"""
A class which encapsulates a process and a set of constraints
and ensures that they are constantly fulfilled.
"""
import datetime
import os
import signal
import simplejson
import sys
import threading
import time

import process


class ProcessHarness(object):
    """
    An object which manages the lifecycle of a single child process, killing it
    when it violates constraitns and rebooting it as necessary
    """
    def __init__(self, command, constraints, restart=False,
                 max_restarts=-1, poll_interval=.1,
                 logmanager=None, uid=None, allow_spam=False,
                 collect_stats=True):
        self.launch_location = os.getcwd()
        self.child_proc = None
        self.child_running = True
        self.collect_stats = collect_stats
        self.command = command
        self.constraints = constraints
        self.max_restarts = max_restarts
        self.poll_interval = poll_interval
        self.restart = restart
        self.start_count = 0
        self.uid = uid
        self.parent_pid = os.getpid()
        self.logmanager = logmanager
        self.logmanager.set_harness(self)
        self.stop_running = False
        self.last_start = datetime.datetime.min
        self.allow_spam = allow_spam
        # Statistics
        self.task_start = datetime.datetime.now()
        self.violations = {}
        for constraint in self.constraints:
            self.violations[str(constraint)] = 0

        signal.signal(signal.SIGTERM, self.exit_now)
        signal.signal(signal.SIGINT, self.exit_now)

        # Start the child process
        self.start_process()

    def exit_now(self, *_):
        self.stop_running = True
        self.child_running = False
        self.terminate_child()
        print simplejson.dumps(self.logmanager.get_logfile_names())
        os._exit(0)

    def start_process(self):
        """
        Start a new instance of the child task
        """
        # Avoid spam-restarts, only allow restarting
        # once per second
        now = datetime.datetime.now()
        if not self.allow_spam and \
                now - self.last_start < datetime.timedelta(seconds=1):
            time.sleep(1)

        self.last_start = datetime.datetime.now()

        pid = os.fork()
        if pid == 0:
            # We're the child, we'll exec
            # Put ourselves into our own pgrp, for sanity
            os.setpgrp()

            if self.uid:
                try:
                    os.setuid(self.uid)
                except OSError:  # no permission
                    sys.stderr.write("Must be root to set UID!")
                    os._exit(1)

            # Configure STDOUT and STDERR
            self.logmanager.setup_stdout()
            self.logmanager.setup_stderr()

            # parse the command
            cmd = '/bin/bash'
            args = [cmd, "-c", self.command]
            os.execvp(cmd, args)

        self.child_proc = process.Process(pid)
        self.start_count += 1

    def do_monitoring(self):
        """
        Begin monitoring the child process
        """
        while True:
            if self.stop_running:
                return

            if self.collect_stats:
                self.child_proc.update_usage(deep=True)

            restarted = False

            for constraint in self.constraints:
                if constraint.check_violation(self.child_proc):
                    if self.child_violation_occured(constraint):
                        print "Restarting child command %s" % self.command
                        self.start_process()
                        restarted = True

            if not self.child_proc.is_alive():
                # The child proc could have died inbetween checking
                # constraints and now.  If there is a LivingConstraint
                # then fire it
                for constraint in self.constraints:
                    if str(constraint) == "LivingConstraint":
                        if self.child_violation_occured(constraint):
                            print "Restarting child command %s" % self.command
                            self.start_process()
                            restarted = True

                # If we restarted the child proc we don't want to set
                # child_running to False because... its True =P
                if not restarted:
                    # There is no living constraint and child is dead,
                    # so set running to false
                    print "Child exited on its own, not asked to " + \
                        "restart it, exiting"
                    self.child_running = False
                    return

            time.sleep(self.poll_interval)

    def child_violation_occured(self, violated_constraint):
        """
        Take appropriate action when we're in violation
        Returns:
          True means the process should be restarted
          False means the caller should take no action
        """
        print "Violated Constraint %s" % str(violated_constraint)
        if violated_constraint.kill_on_violation:
            self.child_proc.force_exit()

        self.violations[str(violated_constraint)] += 1

        if self.restart:
            if (self.max_restarts == -1 or
                self.start_count <= self.max_restarts):
                return True

        self.child_running = False
        return False

    def begin_monitoring(self):
        """Split off a thread to monitor the child process"""
        monitoring_thread = threading.Thread(target=self.do_monitoring,
                                             name='Child Monitoring')

        monitoring_thread.start()

    def terminate_child(self):
        """
        Kill the child process
        """
        self.child_proc.force_exit()
        self.wait_for_child_to_finish()

    def wait_for_child_to_finish(self):
        """
        Wait for the child process to complete naturally.
        """
        code = 0
        while self.child_running:
            _, newcode = self.child_proc.wait_for_completion()
            if newcode:
                code = newcode

            print "Child %s exited %s" % (self.child_proc.pid, code)
            time.sleep(.1)

        return code



# You can set AWS Keys here or in ENV
#os.putenv('AWS_ACCESS_KEY_ID', XXXXXX)
#os.putenv('AWS_SECRET_ACCESS_KEY', XXXXX)

# Location for cerebro logs
log_location = "/mnt/data/clustersitter"

# Keys used to login to machines
# TODO -- should define this again inside provider_config
keys = ["/opt/wifast/keys/WiFastAWSus-west-1.pem",
        "/opt/wifast/keys/WiFastAWSus-west-2.pem",
        "/opt/wifast/keys/WiFastAWSus-east-1.pem"]

# The user to login as in machines created by providers
# TODO -- this should be inside provider_config
login_user = "ubuntu"

# Define configuration for machine providers
provider_config = {
    'aws': {
        'us-east-1a': {
            '32b_image_id': 'ami-8b78afe2',
            '64b_image_id': 'ami-eb915a82',
            'key_name': 'WiFastAWS',
            'security_groups': ['clustersitter'],
            },
        'us-west-2a': {
            '32b_image_id': 'ami-d862efe8',
            '64b_image_id': 'ami-6c15985c',
            'key_name': 'WiFastAWSus-west-2',
            'security_groups': ['clustersitter'],
            },
        'us-west-1a': {
            '32b_image_id': 'ami-7dd48a38',
            '64b_image_id': 'ami-cb8ed48e',
            'key_name': 'WiFastAWSus-west-1',
            'security_groups': ['clustersitter'],
            }
        },
    }

for az in ['b', 'c', 'd']:
    provider_config['aws']['us-east-1%s' % az] = \
        provider_config['aws']['us-east-1a']

    provider_config['aws']['us-west-2%s' % az] = \
        provider_config['aws']['us-west-2a']

    provider_config['aws']['us-west-1%s' % az] = \
        provider_config['aws']['us-west-1a']

# DNS Provider configuration
dns_provider_config = {
    'class': 'dynect:Dynect',
    'customername': '',
    'username': '',
    'password': '',
    'default_domain': ''
    }



"""
A wrapper around optparse to provide argparse compatibility
"""
PARSER = "argparse"

import optparse

try:
    import argparse
except ImportError:
    PARSER = "optparse"


class OptParserWrapper(optparse.OptionParser):
    def __init__(self, description=None, *args, **kwargs):
        optparse.OptionParser.__init__(self, usage=description, *args,
                                     **kwargs)

    def add_argument(self, name, dest=None, helpstr=None, action=None,
                     const=None, default=None, required=None):
        """
        Same args / output as argparse.add_argument
        """
        self.add_option(name, dest=dest, help=helpstr, action=action,
                      default=default)

    def parse_args(self, args):
        """
        Same args / output as argparse.parse_args
        """
        (options, args) = optparse.OptionParser.parse_args(self, args)
        return options


class CustomArgParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        argparse.ArgumentParser.__init__(
            self,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            *args,
            **kwargs)


if PARSER == "argparse":
    ArgumentParser = CustomArgParser
else:
    ArgumentParser = OptParserWrapper



import json
import logging
import time

logger = logging.getLogger(__name__)


class ClusterState(object):
    def __init__(self, parent):
        # list of tuples (MachineMonitor, ThreadObj)
        self.monitors = []
        self.machines_by_zone = {}
        self.zones = []
        self.provider_by_zone = {}
        self.jobs = []
        self.job_fill = {}
        self.job_fill_machines = {}
        self.providers = {}
        self.job_overflow = {}
        self.unreachable_machines = []
        self.machine_spawn_threads = []
        self.idle_machines = []
        self.sitter = parent
        self.job_file = "%s/jobs.json" % self.sitter.log_location
        self.repair_jobs = []
        self.max_idle_per_zone = -1
        self.loggers = []

    def add_job(self, job):
        logger.info("Add Job: %s" % job.name)
        for zone in job.get_shared_fate_zones():
            if not zone in self.zones:
                logger.warn("Tried to add a job with an unknown SFZ %s" % zone)
                return False

        # Ensure we can find the job's recipe
        if (job.deployment_recipe and
            not self.sitter._get_recipe_class(job.deployment_recipe)):
            logger.warn(
                "Tried to add a job with an invalid recipe class: %s" % (
                    job.deployment_recipe))

            return False

        # Ensure we don't already have a job with this name,
        # if we do, replace it
        found_existing_job = False
        for existing_job in self.jobs:
            if existing_job.name == job.name:
                found_existing_job = True
                # Rather than wholesale replace the object we will
                # simply update the 'configured' fields, this way
                # existing state aka job fillers are not lost
                fields = ['dns_basename',
                          'task_configuration',
                          'deployment_layout',
                          'deployment_recipe',
                          'recipe_options',
                          'persistent',
                          'linked_job']
                for field in fields:
                    setattr(existing_job, field, getattr(job, field))

        if not found_existing_job:
            self.jobs.append(job)

        self.persist_jobs()
        return True

    def remove_job(self, jobname):
        removed = False
        for job in self.jobs:
            if job.name == jobname:
                self.jobs.remove(job)
                removed = True

        self.persist_jobs()
        return removed

    def persist_jobs(self):
        """
        Naively just rewrite the whole "job db"
        """
        jobs_to_write = [j for j in self.jobs if j.persistent]
        f = open(self.job_file, 'w')
        f.write(json.dumps([j.to_dict() for j in jobs_to_write]))
        f.close()

    def get_idle_machines_in_zone(self, zone):
        """
        @ TODO Do some sort of calculation -- if we have too many idle
        machines we should decomission them.  Define a configurable
        threshold somewhere.
        """
        return self.idle_machines[zone]

    def calculate_job_fill(self):
        # If we find out that a job has TOO MANY tasks,
        # then we should decomission some machines or make
        # them idle

        job_fill = {}
        job_fill_machines = {}
        #!MACHINEASSUMPTION! Should be cpu_count not machine_count
        # Fill out a mapping of [job][task] -> machine_count
        logger.debug("Calculating job fill for jobs: %s" % self.jobs)
        for job in self.jobs:
            job_fill[job.name] = {}
            job_fill_machines[job.name] = {}

            for zone in job.get_shared_fate_zones():
                job_fill[job.name][zone] = 0
                job_fill_machines[job.name][zone] = []

        # Actually do the counting
        for zone, machines in self.machines_by_zone.items():
            for machine in machines:
                for task in machine.get_running_tasks():
                    # Don't add tasks from machines to the job_fill
                    # dict unless we already know about the job
                    if not task['name'] in job_fill:
                        continue

                    job_fill[task['name']][zone] += 1
                    job_fill_machines[task['name']][zone].append(machine)

        self.job_fill = job_fill
        self.job_fill_machines = job_fill_machines
        logger.debug("Calculated job fill: %s" % self.job_fill)

    def calculate_job_refill(self):
        logger.debug("Calculating job refill for jobs: %s" % self.jobs)
        # Now see if we need to add any new machines to any jobs
        for job in self.jobs:
            if job.name in self.job_fill:
                if not job.refill(self, self.sitter):
                    # refill returning false means
                    # we need to wait for another calculation run before
                    # we can keep working.
                    break

        logger.debug("Calculated job refill: %s" % self.job_fill)

    def calculate_idle_machines(self):
        idle_machines = {}
        for zone in self.zones:
            idle_machines[zone] = []
            for machine in self.machines_by_zone.get(zone, []):
                tasks = machine.get_running_tasks()

                #!MACHINEASSUMPTION! Here we assume no tasks == idle,
                # not sum(jobs.cpu) < machine.cpu etc.
                idle = bool(not tasks)
                idle = idle and machine.has_loaded_data()
                idle = idle and not machine.is_in_deployment()

                if idle:
                    idle_machines[zone].append(machine)

        # The DICT swap must be atomic, or else another
        # thread could get a bad value during calculation.
        self.idle_machines = idle_machines
        logger.debug("Calculated idle machines: %s" % str(self.idle_machines))

    def calculate_job_overfill(self):
        # TODO we really should just do the calculating here
        # and let the machinedoctor spin down the task
        job_overflow = {}
        for job in self.jobs:
            zone_overflow = job.get_zone_overflow(self)
            job_overflow[job.name] = zone_overflow

        self.job_overflow = job_overflow
        logger.debug("Calculated job overflow: %s" % self.job_overflow)

    def _calculate(self):
            def run_job(job, name):
                # Since all state is accessed and shared there
                # are all sorts of race conditions if a calculator
                # is running and a job is added or removed.
                # If one calculator run crashes because of this
                # thats OK.
                try:
                    job()
                except:
                    import traceback
                    logger.warn("Crash in %s" % name)
                    logger.warn(traceback.format_exc())

            run_job(self.calculate_idle_machines, "Calculate Idle Machines")
            run_job(self.calculate_job_fill, "Calculate Job Fill")
            run_job(self.calculate_job_refill, "Calculate Job ReFill")
            run_job(self.calculate_job_overfill, "Calculate Job OverFill")

    def _calculator(self):
        while True:
            self._calculate()
            time.sleep(self.sitter.stats_poll_interval)



class DNSProvider(object):
    def add_record(self, data, hostName, type="A", TTL=3600, domainName=None):
        pass

    def remove_record(self, data, hostName, type="A", domainName=None):
        pass

    def get_records(self, hostName, type="A", domainName=None):
        pass



import logging
import json
import os
import requests
import sys

from sittercommon import arg_parser


def get_help_string():
    return "Update clustersitter job config"


def get_command():
    return "update_job_cfg"


def get_parser(parser):
    parser.add_argument("--job-config", dest="job_config",
                        required=True,
                        help='JSON file of Job configuration')

    return parser


def run_command(clustersitter_url=None,
             job_config=None):
    try:
        jobs = json.load(open(job_config))
    except:
        print "Error opening config file %s" % job_config
        import traceback
        traceback.print_exc()
        return

    print "%s/add_job" % clustersitter_url
    for job in jobs:
        data = {'data': json.dumps(job)}
        resp = requests.post("%s/add_job" % clustersitter_url, data=data)
        print resp.content
